import click
from coreapi.utils import get_installed_codecs

from apistar import schema

codecs = {
    name: codec for name, codec in get_installed_codecs().items()
    if '+' in codec.media_type
}


class Format(schema.String):
    description = 'The format for the API Schema output.'
    default = 'corejson'
    choices = list(codecs.keys())


def schema(format: Format) -> None:  # pragma: nocover
    """
    Output an API Schema.
    """
    from apistar.main import get_current_app
    from apistar.apischema import APISchema
    app = get_current_app()
    schema = APISchema.build(app)
    codec = codecs[format]
    output = codec.encode(schema)
    click.echo(output)



import os
from wsgiref.util import FileWrapper

import apistar
from apistar import exceptions, wsgi
from apistar.compat import whitenoise
from apistar.decorators import exclude_from_schema
from apistar.routing import Path, Route
from apistar.settings import Settings

PACKAGE_DIR = os.path.dirname(apistar.__file__)
PACKAGE_STATICS = os.path.join(PACKAGE_DIR, 'static')


class Statics(object):
    preload = True

    def __init__(self, root_dir=None):
        assert whitenoise is not None, 'whitenoise must be installed.'
        from whitenoise import WhiteNoise
        self.whitenoise = WhiteNoise(application=None, root=root_dir)
        self.whitenoise.add_files(PACKAGE_STATICS, prefix='apistar')

    @classmethod
    def build(cls, settings: Settings):
        root_dir = settings.get(['STATICS', 'DIR'])
        return cls(root_dir)


@exclude_from_schema
def serve_static(path: Path, statics: Statics, environ: wsgi.WSGIEnviron) -> wsgi.WSGIResponse:
    if not path.startswith('/'):
        path = Path('/' + path)
    static_file = statics.whitenoise.files.get(path)
    if static_file is None:
        raise exceptions.NotFound()

    response = static_file.get_response(environ['REQUEST_METHOD'], environ)
    status_line = '{} {}'.format(response.status, response.status.phrase)
    headers = list(response.headers)
    if response.file is not None:
        file_wrapper = environ.get('wsgi.file_wrapper', FileWrapper)
        content = file_wrapper(response.file)
    else:
        # We hit this branch for HEAD requests
        content = []
    return wsgi.WSGIResponse(status_line, headers, content)


static_routes = [
    Route('/{path}', 'GET', serve_static)
]



import base64
import inspect
from typing import Dict, Optional
from urllib.parse import urljoin

import coreschema
from coreapi import Document, Field, Link
from coreapi.codecs import CoreJSONCodec
from uritemplate import URITemplate

from apistar import http, schema
from apistar.app import App
from apistar.decorators import exclude_from_schema
from apistar.routing import (
    Route, RoutesConfig, primitive_types, schema_types, walk
)
from apistar.templating import Templates


class APISchema(Document):
    @classmethod
    def build(cls, app: App, base_url: http.URL=None):
        routes = app.routes
        url = get_schema_url(routes, base_url)
        content = get_schema_content(routes)
        return cls(url=url, content=content)


def get_schema_url(routes: RoutesConfig, base_url: http.URL) -> Optional[str]:
    """
    Given the application routes, return the URL path of the API Schema.
    """
    for route in walk(routes):
        if route.view is serve_schema:
            return urljoin(base_url, route.path)
    return None


def get_schema_content(routes: RoutesConfig) -> Dict[str, Route]:
    """
    Given the application routes, return a dictionary containing all the
    Links that the service exposes.
    """
    content = {}
    for route in walk(routes):
        view = route.view
        if getattr(view, 'exclude_from_schema', False):
            continue
        name = view.__name__
        link = get_link(route)
        content[name] = link
    return content


def get_link(route: Route) -> Link:
    """
    Given a single route, return a Link instance containing all the information
    needed to expose that route in an API Schema.
    """
    path, method, view = route

    view_signature = inspect.signature(view)
    uritemplate = URITemplate(path)

    fields = []
    for param in view_signature.parameters.values():

        if param.annotation is inspect.Signature.empty:
            annotated_type = str
        else:
            annotated_type = param.annotation

        location = None
        required = False
        param_schema = coreschema.String()
        if param.name in uritemplate.variable_names:
            location = 'path'
            required = True
        elif (annotated_type in primitive_types) or issubclass(annotated_type, schema_types):
            if method in ('POST', 'PUT', 'PATCH'):
                if issubclass(annotated_type, schema.Object):
                    location = 'body'
                    required = True
                else:
                    location = 'form'
            else:
                location = 'query'

        if location is not None:
            field = Field(name=param.name, location=location, required=required, schema=param_schema)
            fields.append(field)

    return Link(url=path, action=method, fields=fields)


@exclude_from_schema
def serve_schema(schema: APISchema) -> http.Response:
    codec = CoreJSONCodec()
    content = codec.encode(schema)
    headers = {'Content-Type': codec.media_type}
    return http.Response(content, headers=headers)


@exclude_from_schema
def serve_schema_js(schema: APISchema, templates: Templates) -> http.Response:
    codec = CoreJSONCodec()
    base64_schema = base64.b64encode(codec.encode(schema)).decode('latin1')
    template = templates.get_template('apistar/schema.js')
    content = template.render(base64_schema=base64_schema)
    headers = {'Content-Type': 'application/javascript'}
    return http.Response(content, headers=headers)



import coreschema

from apistar.apischema import APISchema, serve_schema, serve_schema_js
from apistar.decorators import exclude_from_schema
from apistar.http import Path
from apistar.routing import Route
from apistar.templating import Templates


def render_form(link):
    properties = dict([
        (field.name, field.schema or coreschema.String())
        for field in link.fields
    ])
    required = []
    schema = coreschema.Object(properties=properties, required=required)
    return coreschema.render_to_form(schema)


@exclude_from_schema
def serve_docs(schema: APISchema, templates: Templates, path: Path):
    index = templates.get_template('apistar/docs/index.html')
    langs = ['python', 'javascript', 'shell']

    def static(path):
        return '/static/' + path

    def get_fields(link, location):
        return [
            field for field in link.fields
            if field.location == location
        ]

    return index.render(
        document=schema,
        static=static,
        langs=langs,
        get_fields=get_fields,
        render_form=render_form,
        path=path
    )


docs_routes = [
    Route('/', 'GET', serve_docs),
    Route('/schema', 'GET', serve_schema),
    Route('/schema.js', 'GET', serve_schema_js),
]



import os
from typing import List  # noqa

import jinja2

from apistar.exceptions import ConfigurationError
from apistar.pipelines import ArgName
from apistar.settings import Settings


class Templates(jinja2.Environment):
    preload = True

    @classmethod
    def build(cls, settings: Settings):
        template_dirs = settings.get(['TEMPLATES', 'DIRS'])

        package_loaders = [
            jinja2.PrefixLoader({
                'apistar': jinja2.PackageLoader('apistar', 'templates')
            })
        ]  # type: List[jinja2.BaseLoader]

        if template_dirs is None:
            filesystem_loaders = []  # type: List[jinja2.BaseLoader]
        else:
            filesystem_loaders = [
                jinja2.FileSystemLoader(template_dir)
                for template_dir in template_dirs
            ]

        loader = jinja2.ChoiceLoader(package_loaders + filesystem_loaders)
        return Templates(loader=loader)


class Template(jinja2.Template):
    prefix = ''
    suffixes = ['.html', '.txt']
    path_delimiter = '__'

    @classmethod
    def build(cls, arg_name: ArgName, templates: Templates):
        paths = arg_name.split(cls.path_delimiter)
        path = os.path.join(cls.prefix, *paths)
        for suffix in cls.suffixes:
            try:
                return templates.get_template(path + suffix)
            except jinja2.TemplateNotFound:
                pass
        raise ConfigurationError('No template found for "%s".' % arg_name)



# flake8: noqa

try:
    import ujson as json
except ImportError:  # pragma: no cover
    import json  # type: ignore


try:
    import whitenoise
except ImportError:  # pragma: no cover
    whitenoise = None



from apistar import Include, Route
from apistar.docs import docs_routes
from apistar.statics import static_routes
from project.views import welcome

routes = [
    Route('/', 'GET', welcome),
    Include('/docs', docs_routes),
    Include('/static', static_routes)
]



from typing import Union


class ConfigurationError(Exception):
    pass


class SchemaError(Exception):
    def __init__(self, detail: Union[str, dict]) -> None:
        self.detail = detail
        super().__init__(detail)


# Handled exceptions

class APIException(Exception):
    default_status_code = 500
    default_detail = 'Server error'

    def __init__(self,
                 detail: Union[str, dict]=None,
                 status_code: int=None) -> None:
        self.detail = self.default_detail if (detail is None) else detail
        self.status_code = self.default_status_code if (status_code is None) else status_code


class Found(APIException):
    default_status_code = 302
    default_detail = ''

    def __init__(self,
                 location: str,
                 detail: Union[str, dict]=None,
                 status_code: int=None) -> None:
        self.location = location
        super().__init__(detail, status_code)


class ValidationError(APIException):
    default_status_code = 400
    default_detail = 'Invalid request'


class NotFound(APIException):
    default_status_code = 404
    default_detail = 'Not found'


class MethodNotAllowed(APIException):
    default_status_code = 405
    default_detail = 'Method not allowed'


class UnsupportedMediaType(APIException):
    default_status_code = 415
    default_detail = 'Unsupported media type in request'
